@@ -1,23 +1,36 @@
 # -*- coding: utf-8 -*-
-from reportlab.lib.colors import Color, CMYKColor, getAllNamedColors, toColor, \
-    HexColor
-from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
-from reportlab.lib.units import inch, cm
 import base64
-import httplib
+from copy import copy
 import logging
 import mimetypes
 import os.path
 import re
-import reportlab
 import shutil
 import string
 import sys
 import tempfile
-import types
-import urllib
-import urllib2
-import urlparse
+
+import reportlab
+from reportlab.lib.colors import Color, toColor
+from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
+from reportlab.lib.units import inch, cm
+import six
+
+
+try:
+    import httplib
+except ImportError:
+    import http.client as httplib
+
+try:
+    import urllib.request as urllib2
+except ImportError:
+    import urllib2
+
+try:
+    import urllib.parse as urlparse
+except ImportError:
+    import urlparse
 
 # Copyright 2010 Dirk Holtwick, holtwick.it
 #
@@ -33,47 +46,43 @@ import urlparse
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-rgb_re = re.compile("^.*?rgb[(]([0-9]+).*?([0-9]+).*?([0-9]+)[)].*?[ ]*$")
+rgb_re = re.compile(
+    "^.*?rgb[a]?[(]([0-9]+).*?([0-9]+).*?([0-9]+)(?:.*?(?:[01]\.(?:[0-9]+)))?[)].*?[ ]*$")
 
 _reportlab_version = tuple(map(int, reportlab.Version.split('.')))
-if _reportlab_version < (2,1):
+if _reportlab_version < (2, 1):
     raise ImportError("Reportlab Version 2.1+ is needed!")
 
-REPORTLAB22 = _reportlab_version >= (2, 2)
-# print "***", reportlab.Version, REPORTLAB22, reportlab.__file__
-
 log = logging.getLogger("xhtml2pdf")
 
-try:
-    import cStringIO as StringIO
-except:
-    import StringIO
-
 try:
     import PyPDF2
-except:
+except ImportError:
     PyPDF2 = None
 
 try:
     from reportlab.graphics import renderPM
-except:
+except ImportError:
     renderPM = None
 
 try:
     from reportlab.graphics import renderSVG
-except:
+except ImportError:
     renderSVG = None
 
-
-#===============================================================================
+from xhtml2pdf.config.httpconfig import httpConfig
+#=========================================================================
 # Memoize decorator
-#===============================================================================
+#=========================================================================
+
+
 class memoized(object):
+
     """
     A kwargs-aware memoizer, better than the one in python :)
 
-    Don't pass in too large kwargs, since this turns them into a tuple of tuples
-    Also, avoid mutable types (as usual for memoizers)
+    Don't pass in too large kwargs, since this turns them into a tuple of
+    tuples. Also, avoid mutable types (as usual for memoizers)
 
     What this does is to create a dictionnary of {(*parameters):return value},
     and uses it as a cache for subsequent calls to the same method.
@@ -81,18 +90,16 @@ class memoized(object):
     and that are called often. It's a perfect match for our getSize etc...
     """
 
-
     def __init__(self, func):
         self.cache = {}
         self.func = func
-        self.__doc__ = self.func.__doc__ # To avoid great confusion
-        self.__name__ = self.func.__name__ # This also avoids great confusion
-
+        self.__doc__ = self.func.__doc__  # To avoid great confusion
+        self.__name__ = self.func.__name__  # This also avoids great confusion
 
     def __call__(self, *args, **kwargs):
         # Make sure the following line is not actually slower than what you're
         # trying to memoize
-        args_plus = tuple(kwargs.iteritems())
+        args_plus = tuple(six.iteritems(kwargs))
         key = (args, args_plus)
         try:
             if key not in self.cache:
@@ -108,23 +115,72 @@ def ErrorMsg():
     """
     Helper to get a nice traceback as string
     """
-    import traceback, sys
+    import traceback
 
-
-    type = value = tb = limit = None
-    type, value, tb = sys.exc_info()
-    list = traceback.format_tb(tb, limit) + traceback.format_exception_only(type, value)
+    limit = None
+    _type, value, tb = sys.exc_info()
+    _list = traceback.format_tb(tb, limit) + \
+        traceback.format_exception_only(_type, value)
     return "Traceback (innermost last):\n" + "%-20s %s" % (
-        string.join(list[: - 1], ""),
-        list[- 1])
+        " ".join(_list[:-1]),
+        _list[-1])
 
 
 def toList(value):
-    if type(value) not in (types.ListType, types.TupleType):
+    if type(value) not in (list, tuple):
         return [value]
     return list(value)
 
 
+def transform_attrs(obj, keys, container, func, extras=None):
+    """
+    Allows to apply one function to set of keys cheching if key is in container,
+    also trasform ccs key to report lab keys.
+
+    extras = Are extra params for func, it will be call like func(*[param1, param2]) 
+
+    obj = frag
+    keys = [(reportlab, css), ... ]
+    container = cssAttr
+    """
+    cpextras = extras
+
+    for reportlab, css in keys:
+        extras = cpextras
+        if extras is None:
+            extras = []
+        elif not isinstance(extras, list):
+            extras = [extras]
+        if css in container:
+            extras.insert(0, container[css])
+            setattr(obj,
+                    reportlab,
+                    func(*extras)
+                    )
+
+
+def copy_attrs(obj1, obj2, attrs):
+    """
+    Allows copy a list of attributes from object2 to object1.
+    Useful for copy ccs attributes to fragment  
+    """
+    for attr in attrs:
+        value = getattr(obj2, attr) if hasattr(obj2, attr) else None
+        if value is None and isinstance(obj2, dict) and attr in obj2:
+            value = obj2[attr]
+        setattr(obj1, attr, value)
+
+
+def set_value(obj, attrs, value, _copy=False):
+    """
+    Allows set the same value to a list of attributes 
+    """
+    for attr in attrs:
+        if _copy:
+            value = copy(value)
+        setattr(obj, attr, value)
+
+
 @memoized
 def getColor(value, default=None):
     """
@@ -140,7 +196,8 @@ def getColor(value, default=None):
     if value in COLOR_BY_NAME:
         return COLOR_BY_NAME[value]
     if value.startswith("#") and len(value) == 4:
-        value = "#" + value[1] + value[1] + value[2] + value[2] + value[3] + value[3]
+        value = "#" + value[1] + value[1] + \
+            value[2] + value[2] + value[3] + value[3]
     elif rgb_re.search(value):
         # e.g., value = "<css function: rgb(153, 51, 153)>", go figure:
         r, g, b = [int(x) for x in rgb_re.search(value).groups()]
@@ -149,11 +206,10 @@ def getColor(value, default=None):
         # Shrug
         pass
 
-    return toColor(value, default) # Calling the reportlab function
+    return toColor(value, default)  # Calling the reportlab function
 
 
 def getBorderStyle(value, default=None):
-    # log.debug(value)
     if value and (str(value).lower() not in ("none", "hidden")):
         return value
     return default
@@ -211,11 +267,11 @@ def getSize(value, relative=0, base=None, default=0.0):
         original = value
         if value is None:
             return relative
-        elif type(value) is types.FloatType:
+        elif type(value) is float:
             return value
         elif isinstance(value, int):
             return float(value)
-        elif type(value) in (types.TupleType, types.ListType):
+        elif type(value) in (tuple, list):
             value = "".join(value)
         value = str(value).strip().lower().replace(",", ".")
         if value[-2:] == 'cm':
@@ -229,21 +285,25 @@ def getSize(value, relative=0, base=None, default=0.0):
         elif value[-2:] == 'pt':
             return float(value[:-2].strip())
         elif value[-2:] == 'pc':
-            return float(value[:-2].strip()) * 12.0 # 1pc == 12pt
+            return float(value[:-2].strip()) * 12.0  # 1pc == 12pt
         elif value[-2:] == 'px':
-            return float(value[
-                         :-2].strip()) * dpi96  # XXX W3C says, use 96pdi http://www.w3.org/TR/CSS21/syndata.html#length-units
+            # XXX W3C says, use 96pdi
+            # http://www.w3.org/TR/CSS21/syndata.html#length-units
+            return float(value[:-2].strip()) * dpi96
         elif value[-1:] == 'i':  # 1pt == 1/72inch
             return float(value[:-1].strip()) * inch
         elif value in ("none", "0", "auto"):
             return 0.0
         elif relative:
             if value[-2:] == 'em':  # XXX
-                return float(value[:-2].strip()) * relative  # 1em = 1 * fontSize
-            elif value[-2:] == 'ex': # XXX
-                return float(value[:-2].strip()) * (relative / 2.0) # 1ex = 1/2 fontSize
+                # 1em = 1 * fontSize
+                return float(value[:-2].strip()) * relative
+            elif value[-2:] == 'ex':  # XXX
+                # 1ex = 1/2 fontSize
+                return float(value[:-2].strip()) * (relative / 2.0)
             elif value[-1:] == '%':
-                return (relative * float(value[:-1].strip())) / 100.0 # 1% = (fontSize * 1) / 100
+                # 1% = (fontSize * 1) / 100
+                return (relative * float(value[:-1].strip())) / 100.0
             elif value in ("normal", "inherit"):
                 return relative
             elif value in _relativeSizeTable:
@@ -258,7 +318,7 @@ def getSize(value, relative=0, base=None, default=0.0):
                 return max(MIN_FONT_SIZE, relative * float(value))
         try:
             value = float(value)
-        except:
+        except ValueError:
             log.warn("getSize: Not a float %r", value)
             return default  # value = 0
         return max(0, value)
@@ -280,7 +340,7 @@ def getCoords(x, y, w, h, pagesize):
         x = ax + x
     if y < 0:
         y = ay + y
-    if w != None and h != None:
+    if w is not None and h is not None:
         if w <= 0:
             w = (ax - x + w)
         if h <= 0:
@@ -299,7 +359,7 @@ def getBox(box, pagesize):
     """
     box = str(box).split()
     if len(box) != 4:
-        raise Exception, "box not defined right way"
+        raise Exception("box not defined right way")
     x, y, w, h = [getSize(pos) for pos in box]
     return getCoords(x, y, w, h, pagesize)
 
@@ -349,7 +409,7 @@ def getPos(position, pagesize):
     """
     position = str(position).split()
     if len(position) != 2:
-        raise Exception, "position not defined right way"
+        raise Exception("position not defined right way")
     x, y = [getSize(pos) for pos in position]
     return getCoords(x, y, None, None, pagesize)
 
@@ -385,15 +445,16 @@ GAE = "google.appengine" in sys.modules
 
 if GAE:
     STRATEGIES = (
-        StringIO.StringIO,
-        StringIO.StringIO)
+        six.BytesIO,
+        six.BytesIO)
 else:
     STRATEGIES = (
-        StringIO.StringIO,
+        six.BytesIO,
         tempfile.NamedTemporaryFile)
 
 
 class pisaTempFile(object):
+
     """
     A temporary file implementation that uses memory unless
     either capacity is breached or fileno is requested, at which
@@ -421,7 +482,7 @@ class pisaTempFile(object):
         self.strategy = int(len(buffer) > self.capacity)
         try:
             self._delegate = self.STRATEGIES[self.strategy]()
-        except:
+        except IndexError:
             # Fallback for Google AppEnginge etc.
             self._delegate = self.STRATEGIES[0]()
         self.write(buffer)
@@ -430,7 +491,8 @@ class pisaTempFile(object):
 
     def makeTempFile(self):
         """
-        Switch to next startegy. If an error occured stay with the first strategy
+        Switch to next startegy. If an error occured,
+        stay with the first strategy
         """
 
         if self.strategy == 0:
@@ -461,14 +523,18 @@ class pisaTempFile(object):
 
     def getvalue(self):
         """
-        Get value of file. Work around for second strategy
+        Get value of file. Work around for second strategy.
+        Always returns bytes
         """
 
         if self.strategy == 0:
             return self._delegate.getvalue()
         self._delegate.flush()
         self._delegate.seek(0)
-        return self._delegate.read()
+        value = self._delegate.read()
+        if not isinstance(value, six.binary_type):
+            value = value.encode('utf-8')
+        return value
 
     def write(self, value):
         """
@@ -485,6 +551,10 @@ class pisaTempFile(object):
                     (self.tell() + len_value) >= self.capacity
             if needs_new_strategy:
                 self.makeTempFile()
+
+        if not isinstance(value, six.binary_type):
+            value = value.encode('utf-8')
+
         self._delegate.write(value)
 
     def __getattr__(self, name):
@@ -497,10 +567,12 @@ class pisaTempFile(object):
             raise AttributeError(e)
 
 
-_rx_datauri = re.compile("^data:(?P<mime>[a-z]+/[a-z]+);base64,(?P<data>.*)$", re.M | re.DOTALL)
+_rx_datauri = re.compile(
+    "^data:(?P<mime>[a-z]+/[a-z]+);base64,(?P<data>.*)$", re.M | re.DOTALL)
 
 
 class pisaFileObject:
+
     """
     XXX
     """
@@ -513,14 +585,16 @@ class pisaFileObject:
         self.uri = None
         self.local = None
         self.tmp_file = None
-        uri = uri.encode('utf-8')
+        uri = uri or str()
+        if type(uri) != str:
+            uri = uri.decode("utf-8")
         log.debug("FileObject %r, Basepath: %r", uri, basepath)
 
         # Data URI
         if uri.startswith("data:"):
             m = _rx_datauri.match(uri)
             self.mimetype = m.group("mime")
-            self.data = base64.decodestring(m.group("data"))
+            self.data = base64.b64decode(m.group("data").encode("utf-8"))
 
         else:
             # Check if we have an external scheme
@@ -529,59 +603,70 @@ class pisaFileObject:
             else:
                 urlParts = urlparse.urlparse(uri)
 
-            log.debug("URLParts: %r", urlParts)
+            log.debug("URLParts: {}".format((urlParts, urlParts.scheme)))
 
             if urlParts.scheme == 'file':
                 if basepath and uri.startswith('/'):
                     uri = urlparse.urljoin(basepath, uri[1:])
                 urlResponse = urllib2.urlopen(uri)
-                self.mimetype = urlResponse.info().get("Content-Type", '').split(";")[0]
+                self.mimetype = urlResponse.info().get(
+                    "Content-Type", '').split(";")[0]
                 self.uri = urlResponse.geturl()
                 self.file = urlResponse
 
-            # Drive letters have len==1 but we are looking for things like http:
+            # Drive letters have len==1 but we are looking
+            # for things like http:
             elif urlParts.scheme in ('http', 'https'):
 
+                log.debug("Sending request for {} with httplib".format(uri))
+
                 # External data
                 if basepath:
                     uri = urlparse.urljoin(basepath, uri)
 
+                log.debug("Uri parsed: {}".format(uri))
+
                 #path = urlparse.urlsplit(url)[2]
                 #mimetype = getMimeType(path)
 
                 # Using HTTPLIB
-                server, path = urllib.splithost(uri[uri.find("//"):])
+                server, path = urllib2.splithost(uri[uri.find("//"):])
                 if uri.startswith("https://"):
-                    conn = httplib.HTTPSConnection(server)
+                    conn = httplib.HTTPSConnection(server,  **httpConfig)
                 else:
                     conn = httplib.HTTPConnection(server)
                 conn.request("GET", path)
                 r1 = conn.getresponse()
                 # log.debug("HTTP %r %r %r %r", server, path, uri, r1)
                 if (r1.status, r1.reason) == (200, "OK"):
-                    self.mimetype = r1.getheader("Content-Type", '').split(";")[0]
+                    self.mimetype = r1.getheader(
+                        "Content-Type", '').split(";")[0]
                     self.uri = uri
+                    log.debug("here")
                     if r1.getheader("content-encoding") == "gzip":
                         import gzip
-                        try:
-                            import cStringIO as StringIO
-                        except:
-                            import StringIO
 
-                        self.file = gzip.GzipFile(mode="rb", fileobj=StringIO.StringIO(r1.read()))
+                        self.file = gzip.GzipFile(
+                            mode="rb", fileobj=six.StringIO(r1.read()))
                     else:
                         self.file = r1
                 else:
+                    log.debug(
+                        "Received non-200 status: {}".format((r1.status, r1.reason)))
                     try:
                         urlResponse = urllib2.urlopen(uri)
-                    except urllib2.HTTPError:
+                    except urllib2.HTTPError as e:
+                        log.error("Could not process uri: {}".format(e))
                         return
-                    self.mimetype = urlResponse.info().get("Content-Type", '').split(";")[0]
+                    self.mimetype = urlResponse.info().get(
+                        "Content-Type", '').split(";")[0]
                     self.uri = urlResponse.geturl()
                     self.file = urlResponse
 
             else:
 
+                log.debug("Unrecognized scheme, assuming local file path")
+
                 # Local data
                 if basepath:
                     uri = os.path.normpath(os.path.join(basepath, uri))
@@ -589,8 +674,13 @@ class pisaFileObject:
                 if os.path.isfile(uri):
                     self.uri = uri
                     self.local = uri
+
                     self.setMimeTypeByName(uri)
-                    self.file = open(uri, "rb")
+                    if self.mimetype and self.mimetype.startswith('text'):
+                        self.file = open(uri, "r") #removed bytes... lets hope it goes ok :/
+                    else:
+                        # removed bytes... lets hope it goes ok :/
+                        self.file = open(uri, "rb")
 
     def getFile(self):
         if self.file is not None:
@@ -617,7 +707,14 @@ class pisaFileObject:
         if self.data is not None:
             return self.data
         if self.file is not None:
-            self.data = self.file.read()
+            try:
+                self.data = self.file.read()
+            except:
+                if self.mimetype and self.mimetype.startswith('text'):
+                    self.file = open(self.file.name, "rb") #removed bytes... lets hope it goes ok :/
+                    self.data = self.file.read().decode('utf-8')
+                else:
+                    raise
             return self.data
         return None
 
@@ -815,4 +912,3 @@ COLOR_BY_NAME = {
     'yellow': Color(1, 1, 0),
     'yellowgreen': Color(.603922, .803922, .196078)
 }
-
